@@ -26,7 +26,6 @@ from demyst.common.pii import PII_REGEX
 from demyst.analytics.shiny import progressbar
 from demyst.analytics.shiny import generatehandlerbutton
 from demyst.analytics.report import report
-from demyst.analytics.type_guesser import TypeGuesser
 
 import os
 import pandas as pd
@@ -582,6 +581,7 @@ class Analytics(object):
         return list(map(lambda provider: provider['name'], filtered_providers))
 
     def input_file_headers(self, table_name):
+        """Returns information about the columns of an input file."""
         return track_function(lambda: self.__do_input_file_headers(table_name),
                               self.__config, "API : Python : Rosefish Product Headers")
 
@@ -596,7 +596,8 @@ class Analytics(object):
             print("Warning: could not find input file: {}".format(table_name), file=sys.stderr)
 
     def list_stats_products(self):
-         return track_function(lambda: self.__do_list_stats_products(),
+        """Returns a list of all data products for which stats are available."""
+        return track_function(lambda: self.__do_list_stats_products(),
                               self.__config, "API : Python : Hogfish Product List")
 
     def __do_list_stats_products(self):
@@ -731,6 +732,7 @@ class Analytics(object):
         return(res.json())
 
     def data_function_append(self, name, inputs_dframe):
+        """Calls a data function and augments the input dataframe with returned results."""
         return track_function(lambda: self.__do_data_function_append(name, inputs_dframe),
                               self.__config, "API : Python : Data Function Append",
                               { "name": name })
@@ -744,8 +746,7 @@ class Analytics(object):
             output[idx] = inputs
             post = {
                 "inputs": inputs,
-                "providers": [],
-                "data_function_id": name
+                "providers": [name]
             }
             channel_url = self.__config.get("BLACKFIN_URL") + "v2/execute"
             resp = self.__config.auth_post(channel_url,
@@ -760,29 +761,8 @@ class Analytics(object):
         sorted_columns = list(inputs_dframe.columns) + [x for x in appended_dframe.columns if x not in list(inputs_dframe.columns)]
         return appended_dframe[sorted_columns]
 
-    # Use this to fetch sample data for a list of providers.  The
-    # input dataframe is validated but ignored, it should have (at
-    # least) one row.
-    def fetch_sample_data(self, providers, inputs_dframe):
-        row = inputs_dframe.iloc[0,:] # Only need one row
-        inputs = dict([(key, row[key]) for key in row.keys()])
-        no_failures = self.__connectors.fetch(providers, inputs=inputs, sample_mode=True)
-        if not no_failures:
-            for provider_name in providers:
-                provider_error = self.__connectors.cache_get_error(provider_name, inputs)
-                if provider_error:
-                    print("Transaction with %s failed because of %s: %s. "%(provider_name, provider_error['type'], provider_error['message']), file=sys.stderr)
-        # Concat results
-        df = inputs_dframe.head(1)
-        for provider_name in providers:
-            provider_data = self.__connectors.get(provider_name, '', inputs=inputs, shape="table", prefix=True)
-            if provider_data:
-                provider_df = pd.DataFrame.from_dict([provider_data])
-                df = pd.concat([df, provider_df], axis="columns")
-        return df
-
     # Easy-to-use frontend to enrich, enrich_wait, and enrich_download
-    def enrich_and_download(self, providers, inputs=None, hosted_input=None, validate=True, all_updates=False, fixed_list_size=None):
+    def enrich_and_download(self, providers, inputs=None, hosted_input=None, validate=True, all_updates=False, fixed_list_size=1):
         """
         Augments your input data with results from our data providers.
         This is the main entry point to the Demyst data platform.
@@ -795,15 +775,7 @@ class Analytics(object):
                          fixed_list_size=fixed_list_size)
         return self.enrich_download(id)
 
-    def enrich_with_hosted_inputs(self, providers, hosted_input, row_limit=None, filters=None):
-        """
-        Like enrich_and_download(), but instead of enriching an input
-        dataframe, it uses sample inputs (see input_files()).
-        """
-        inputs = self.input_file(hosted_input, row_limit=row_limit, filters=filters)
-        return self.enrich_and_download(providers=providers, inputs=inputs, validate=False)
-
-    def enrich(self, providers, inputs=None, hosted_input=None, validate=True, all_updates=False, fixed_list_size=None):
+    def enrich(self, providers, inputs=None, hosted_input=None, validate=True, all_updates=False, fixed_list_size=1):
         """
         enrich() is the lower-level (compared to enrich_and_download())
         workhorse that lets you kick off an enrichment job asynchronously.
@@ -832,7 +804,7 @@ class Analytics(object):
         path = "table_providers/calculate_credit_cost?" + urlencode(url_params, doseq=True)
         return self.__config.auth_get(self.__config.get("MANTA_URL") + path).json()["total_cost"]
 
-    def __do_enrich(self, providers, inputs, validate=True, all_updates=False, fixed_list_size=None):
+    def __do_enrich(self, providers, inputs, validate=True, all_updates=False, fixed_list_size=1):
         limit = 20
         if len(providers) > limit:
             raise RuntimeError("Only " + str(limit) + " providers are allowed.")
@@ -1047,21 +1019,22 @@ class Analytics(object):
             messages.append(name + " (" + status + ")")
         return "<br>".join(messages)
 
-    def enrich_download(self, id, block_until_complete=True):
+    def enrich_download(self, id, providers=[], block_until_complete=True):
         """
         Downloads the augumented data of an enrichment job and returns
         the resulting dataframe.
         """
-        return track_function(lambda: self.__do_enrich_download(id, block_until_complete=block_until_complete),
+        return track_function(lambda: self.__do_enrich_download(id, providers=providers, block_until_complete=block_until_complete),
                               self.__config, "API : Python : Download",
                               { "job_id": id })
 
-    def __do_enrich_download(self, id, block_until_complete=True):
+    def __do_enrich_download(self, id, providers=[], block_until_complete=True):
         fd, temp_path = tempfile.mkstemp()
         try:
             # Overwrite=True is needed because temp file gets created above
             self.enrich_download_to_disk(id,
                                          temp_path,
+                                         providers=providers,
                                          overwrite=True,
                                          block_until_complete=block_until_complete)
             return pd.read_csv(temp_path, na_filter=False)
@@ -1073,7 +1046,7 @@ class Analytics(object):
                 #print("Warning: couldn't delete temporary file: " + temp_path, file=sys.stderr)
                 pass
 
-    def enrich_download_to_disk(self, id, file_path, overwrite=False, block_until_complete=True):
+    def enrich_download_to_disk(self, id, file_path, providers=[], overwrite=False, block_until_complete=True):
         """
         Downloads the augumented data of an enrichment job and saves
         it as a CSV file on disk.
@@ -1091,11 +1064,12 @@ class Analytics(object):
             export_link = r["most_recent_export_link"]
             if export_link:
                 provider_name = r["table_provider"]["name"]
-                temp_fd, temp_path = tempfile.mkstemp(prefix=provider_name, suffix=".csv")
-                exports.append({ "provider_name": provider_name,
-                                 "export_link": export_link,
-                                 "temp_fd": temp_fd,
-                                 "temp_path": temp_path })
+                if (len(providers) == 0) or (provider_name in providers):
+                    temp_fd, temp_path = tempfile.mkstemp(prefix=provider_name, suffix=".csv")
+                    exports.append({ "provider_name": provider_name,
+                                     "export_link": export_link,
+                                     "temp_fd": temp_fd,
+                                     "temp_path": temp_path })
             else:
                 print("Provider " +provider_name+ " has not finished enrichment and will not be included." , file=sys.stderr)
         # Download all export links to temp files in parallel
@@ -1224,6 +1198,10 @@ class Analytics(object):
         else:
             for connector in connectors:
                 connector_inputs = list(set(list(itertools.chain(*connector['required_input_name_sets']))))
+                if 'optional_input_name_sets' in connector:
+                    connector_inputs_opt = list(set(list(itertools.chain(*connector['optional_input_name_sets']))))
+                else:
+                    connector_inputs_opt = []
                 with tag('div', style=''):
                     with tag('div', style="pading:10px;"):
                         with tag('div',style='padding:5px'):
@@ -1234,6 +1212,7 @@ class Analytics(object):
                             with tag('p'):
                                 if connector['description']:
                                     line('em', connector['description'])
+                        # Required inputs
                         with tag('div', style='padding: 5px;'):
                             with tag('table'):
                                 with tag('tr'):
@@ -1256,6 +1235,29 @@ class Analytics(object):
                                                             text("☐")
                                                     else:
                                                         text(" ")
+                            # Optional inputs
+                            if len(connector_inputs_opt) > 0:
+                                with tag('table'):
+                                    with tag('tr'):
+                                        with tag('th'):
+                                            text(" ")
+                                            for i in range(len(connector_inputs_opt)):
+                                                with tag('th'):
+                                                    text(connector_inputs_opt[i])
+                                    for ri_idx, opt_inputs in enumerate(connector['optional_input_name_sets']):
+                                        with tag('tr'):
+                                            with tag('td'):
+                                                text("Optional inputs")
+                                            for i in range(len(connector_inputs_opt)):
+                                                with tag('td'):
+                                                    with tag('span', style="font-size: 20px;"):
+                                                        if connector_inputs_opt[i] in opt_inputs:
+                                                            if inputs and (connector_inputs_opt[i] in inputs):
+                                                                text("☒")
+                                                            else:
+                                                                text("☐")
+                                                        else:
+                                                            text(" ")
         return HTML(doc.getvalue())
 
     def __filter_connectors_strictly(self, connectors, inputs):
@@ -1318,15 +1320,21 @@ class Analytics(object):
             return validation.all_valid()
 
     def credits(self):
+        """Returns the number of credits that are available for running enrichments."""
         headers = { "Accept": "application/json" }
         resp_json = self.__config.auth_get(self.__config.get("MANTA_URL") + "organization", headers=headers).json()
         balance = resp_json["credit_balance"]
         return balance
 
+    HIDE_SETTINGS=["MANTA_URL", "BLACKFIN_URL", "PASSWORD"]
+    
     # Print configuration settings
     def show_settings(self):
         for (key, val) in self.__config.settings.items():
-            print(key + ": " + val)
+            if (not key in self.HIDE_SETTINGS):
+                print(key.title() + ": " + val)
+        print("Organization: " + str(self.__config.get_organization_name()))
+        print("Organization ID: " + str(self.__config.get_organization()))
 
     # Product catalog: simple JSON info about available providers
     def products(self, params=None, full=False, dataframe=True):
@@ -1420,12 +1428,16 @@ class Analytics(object):
     # between search(view="html") and product_inputs(view="html")
     def __transform_provider_to_connector(self, provider):
         required_input_name_sets = []
+        optional_input_name_sets = []
         if type(provider["version"]) is dict:
             version = provider["version"]
             for data in version["schema"]:
                 for variant in data["input"]["required"]:
-                    name_set = [input["name"] for input in variant]
-                    required_input_name_sets.append(name_set)
+                    req_name_set = [input["name"] for input in variant]
+                    required_input_name_sets.append(req_name_set)
+                if "optional" in data["input"]:
+                    opt_name_set = [input["name"] for input in data["input"]["optional"]]
+                    optional_input_name_sets.append(opt_name_set)
         provider_name = provider.get("name") or ""
         data_source = provider.get("aegean_data_source")
         if data_source:
@@ -1434,6 +1446,7 @@ class Analytics(object):
             data_source_name = provider_name
         return {
             "required_input_name_sets": required_input_name_sets,
+            "optional_input_name_sets": optional_input_name_sets,
             "data_source_logo_url": provider.get("logo") or "",
             "data_source_name": data_source_name,
             "name": provider_name,
@@ -1510,6 +1523,7 @@ class Analytics(object):
 
     def load_input(self, csv_path_or_df):
         """Automatically infers the type of input from the CSV file."""
+        from demyst.analytics.type_guesser import TypeGuesser
         t = TypeGuesser(csv_path_or_df)
         t.analyze()
         return(t.data.astype(str))
@@ -1738,6 +1752,7 @@ class Analytics(object):
         return(params)
 
     def tracking(self, start=None, end=None, limit=None):
+        """Returns a list of transactions between the given start date and end date."""
         start_date, end_date = self.__calculcate_tracking_dates(start, end, datetime.utcnow())
         limit = limit or self.__default_tracking_limit()
         recent_params = {
@@ -1779,6 +1794,80 @@ class Analytics(object):
     def __default_tracking_limit(self):
         return 50
 
+    # Errors
+
+    RESOLUTIONS = {
+        "unexpected_upstream_structure": "Contact Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "upstream_application_error": "Contact Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "data_function_misses_result_key": "Contact Demyst Platform support team with data function name",
+        "insufficient_credentials": "Contact Demyst Data team with the provider name",
+        "upstream_service_unavailable": "Re-try after some time and if the issue persists, contact the Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "unexpected_upstream_http_status": "Re-try after some time and if the issue persists, contact the Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "invalid_credentials": "Contact Demyst Data team with the provider name",
+        "rate_limit_exceeded": "Contact Demyst Data team with the provider name",
+        "logical_input_error": "Re-check the input values. Use the validate method to identify errors with the input values.",
+        "timeout": "Re-try after some time and if the issue persists, contact the Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "unexpected_provider_error": "Contact Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "network_error": "Re-try after some time and if the issue persists, contact the Demyst Integration support team with the provider, enrich job id, and input records (if necessary)",
+        "insufficient_input": "Use validate() to check your inputs",
+        "validation_error": "Re-check the input values. An empty string or a blank space could cause this error.",
+        "insufficient_api_key_permissions": "Re-run after authorizing with the right API key",
+        }
+    
+    # Attempts to parse error message and return tuple (type, msg).
+    # If it cannot be parsed returns ("other_error", ...) where ... is
+    # the whole error message.
+    def __parse_error_message(self, full_msg):
+        parts = [part.strip() for part in full_msg.split(",")]
+        type = None
+        msg = None
+        for p in parts:
+            key, val = [s.strip() for s in p.split(":")]
+            if key == "type":
+                type = val
+            elif key == "message":
+                msg = val
+        if type and msg:
+            return (type, msg)
+        else:
+            return ("other_error", full_msg)
+
+    def __find_error_resolution(self, error_type):
+        return self.RESOLUTIONS.get(error_type) or ""
+
+    def enrich_errors(self, df):
+        errors = {} # provider + "." + type -> { count, message }
+        error_columns = [col for col in list(df) if col.endswith(".error")]
+        for index, row in df.iterrows():
+            for col in error_columns:
+                if (row[col] != "") and (pd.notna(row[col])):
+                    re_product_name = re.search('^([^\\.]*)\\.(.*)$', col)
+                    product_name = re_product_name.group(1)
+                    type, msg = self.__parse_error_message(row[col])
+                    error = errors.get(product_name + "." + type)
+                    if not error:
+                        error = { "count": 0, "message": msg }
+                        errors[product_name + "." + type] = error
+                    error["count"] = error["count"] + 1
+                    error["message"] = msg
+        results = []
+        for key, error in errors.items():
+            re_product_name = re.search('^([^\\.]*)\\.(.*)$', key)
+            product_name = re_product_name.group(1)
+            type = re_product_name.group(2)
+            row = {
+                "Provider": product_name,
+                "Error Type": type,
+                "Error Message": error["message"],
+                "Count": error["count"],
+                "Percentage": "{:.2f}%".format(100 * float(error["count"]) / float(len(df))),
+                "Resolution": self.__find_error_resolution(type)
+                }
+            results.append(row)
+        results_df = pd.DataFrame(results)
+        results_df = results_df.sort_values(["Provider", "Error Type"]).reset_index()
+        return results_df[["Provider", "Error Type", "Error Message", "Count", "Percentage", "Resolution"]]
+    
 def show_package_versions():
     installed_packages = pkg_resources.working_set
     for i in installed_packages:
